from .base_sql_connector import BaseSqlConnector
from cnvrg.modules.errors import CnvrgError
try:
    import snowflake
    import snowflake.connector
    imported = True
except ImportError:
    imported = False



class SnowflakeConnector(BaseSqlConnector):

    @staticmethod
    def key_type():
        return "snowflake"

    @staticmethod
    def requirements():
        return """snowflake"""

    def __init__(self, data_connector):
        if not imported: raise CnvrgError("Cant find module snowflakes")
        super(SnowflakeConnector, self).__init__(data_connector)

    def _connection(self):
        return snowflake.connector.connect(**self.data)


import click
from cnvrg.helpers.auth_helper import CnvrgCredentials
import cnvrg.helpers.logger_helper as logger_helper
from tinynetrc import Netrc

@click.group()
def cnvrg():
    pass

@cnvrg.command()
@click.option("--email")
@click.option("--password")
@click.option("--api")
def login(email=None, password=None, api=None):
    cred = CnvrgCredentials()
    if cred.logged_in: return logger_helper.log_message("Already logged in", level=logger_helper.LEVEL_SUCCESS)
    if not email: email = click.prompt("Please enter your email: ")
    if not password: password = click.prompt("Please enter your password: (hidden)", hide_input=True)
    try:
        cred.login(email, password, api_url=api)
        logger_helper.log_message("Logged in as {username}".format(username=cred.username))
    except Exception as e:
        print(e)
        logger_helper.log_error(e)
        logger_helper.log_message("Cant log in, please check your input params", level=logger_helper.LEVEL_ERROR)

@cnvrg.command()
@click.option("--email")
@click.option("--token")
@click.option("--api")
def token_login(email=None, token=None, api=None):
    cred = CnvrgCredentials()
    if cred.logged_in: return logger_helper.log_message("Already logged in", level=logger_helper.LEVEL_SUCCESS)
    if not email: email = click.prompt("Please enter your email: ")
    if not token: token = click.prompt("Please enter your token: (hidden)", hide_input=True)
    try:
        cred.token_login(email, token, api_url=api)
        logger_helper.log_message("Logged in as {username}".format(username=cred.username))
    except Exception as e:
        print(e)
        logger_helper.log_error(e)
        logger_helper.log_message("Cant log in, please check your input params", level=logger_helper.LEVEL_ERROR)



@cnvrg.command()
def logout():
    cred = CnvrgCredentials()
    if cred.logout(): logger_helper.log_message("Logged out successfuly", logger_helper.LEVEL_SUCCESS)
    else: logger_helper.log_message("You are already logged out", logger_helper.LEVEL_SUCCESS)


import random
import string
def random_string(string_length):
    """Generate a random string with the combination of lowercase and uppercase letters """
    letters = string.ascii_letters
    return ''.join(random.choice(letters) for i in range(string_length))


import os
import time
import json
import types
from croniter import croniter
import numpy as np

from enum import Enum
from typing import List

from cnvrg.modules.project import Project
from cnvrg.modules.errors import UserError
from cnvrg.modules.errors import CnvrgError, APINotFound
from cnvrg.helpers.async_runner import AsyncRunner
from cnvrg.helpers.env_helper import in_experiment
from cnvrg.helpers.url_builder_helper import url_join
from cnvrg.helpers.libs_helper import check_if_lib_exists
from cnvrg.modules.mixins.charts_mixin import ChartsMixin
from cnvrg.helpers.error_catcher import suppress_exception
from cnvrg.helpers.param_helper import wrap_string_to_list
from cnvrg.modules.cnvrg_job import CnvrgJob, LOGS_TYPE_ERROR

import cnvrg.helpers.env_helper as env_helper
import cnvrg.helpers.apis_helper as apis_helper
import cnvrg.helpers.spawn_helper as spawn_helper
import cnvrg.helpers.logger_helper as logger_helper
import cnvrg.helpers.string_helper as string_helper
import cnvrg.helpers.chart_show_helper as chart_show_helper
import cnvrg.helpers.param_build_helper as param_build_helper
import cnvrg.helpers.control_stdout_helper as control_stdout_helper


class TagType(Enum):
    SINGLE_TAG = "single"
    LINECHART_TAG = "linechart"


class Experiment(CnvrgJob, ChartsMixin):
    def __init__(self, experiment=None, owner=None):
        default_owner, project_slug, slug = param_build_helper.parse_params(experiment, param_build_helper.EXPERIMENT)
        owner = owner if owner is not None else default_owner
        if not in_experiment() and not slug:
            raise UserError("Cant create an experiment without slug and outside experiment run")
        slug = slug or env_helper.get_current_job_id()
        super(Experiment, self).__init__(slug, env_helper.EXPERIMENT, Project(url_join(owner, project_slug)))
        self.__data = self.__get_experiment()
        self.chart_metrics = {}


    @staticmethod
    @suppress_exception
    def _create_clean_experiment(project=None, **kwargs):
        if isinstance(project, str):
            project = Project(project)
        project = project or Project()
        kwargs["clean_experiment"] = True
        resp = apis_helper.post(url_join(project.get_base_url(), 'experiments', 'local'),
                                data={"commit": project.get_current_commit(), **kwargs, **env_helper.get_origin_job()})
        e = resp.get("experiment")
        exp = Experiment(url_join(project.get_project_name(), e.get("slug")), owner=project.owner)
        exp.notify_running()
        exp.set_as_default()
        return exp

    @staticmethod
    def init(project=None, **kwargs):
        return Experiment._create_clean_experiment(project=project, **kwargs)

    def notify_running(self):
        logger_helper.log_message(
            "{title} is running, follow live at: {url}".format(title=self['title'], url=self.href()))

    def set_as_default(self):
        # If we are already in the context of another job - don't override (for running experiments from notebooks)
        if env_helper.get_origin_job() and env_helper.get_current_job_type != "Experiment":
            return
        env_helper.set_current_job_id(self.job_slug)
        env_helper.set_current_job_type("Experiment")

    @staticmethod
    def __run_callable(function_command, title=None, project=None, library=None, **kwargs):
        experiment = Experiment._create_clean_experiment(title=title, project=project, library=library, **kwargs)
        exit_status = control_stdout_helper.run_callable(
            function_command,
            arguments=[experiment],
            callback=lambda x: experiment.log(x),
            err_callback=lambda x: experiment.log(x, log_type=LOGS_TYPE_ERROR)
        )
        experiment.finish(exit_status=exit_status)
        return experiment

    @staticmethod
    def __run_local(cmd, **kwargs):
        def function_command(experiment=None):
            env = None
            if experiment:
                env = experiment.as_env()
            return spawn_helper.run_sync(cmd, print_output=True, cwd=kwargs["cwd"], env=env)["code"]
        return Experiment.__run_callable(function_command, cmd=cmd, **kwargs)

    def as_env(self):
        return {
            env_helper.ENV_KEYS["current_job_id"]: self.job_slug,
            env_helper.ENV_KEYS["current_job_type"]: "Experiment",
            env_helper.ENV_KEYS["current_project"]: self.project.slug,
            env_helper.ENV_KEYS["current_organization"]: self.project.owner,
        }

    def sync(self, output_dir='output', git_diff=False):
        payload = {
            "job_type": "Experiment",
            "job": self.job_slug,
            "in_exp": "true"
        }

        if self.__data["compute"] == "local":
            payload["local"] = "true"

        if self.project.is_git():
            payload["output_dir"] = output_dir
            if git_diff:
                payload["git_diff"] = git_diff

        self.project.sync(**payload)

    def log_images(self, files: list, target=None):
        try:
            filters = ['image']
            payload = {
                "job_type": "Experiment",
                "workflow_slug": self.job_slug,
                "source": 'sdk'
            }
            self.project.put_files(files, message="Log Images Commit", payload=payload, prefix=target, filters=filters, tag_images=True)
        except APINotFound:  # To support old server version
            payload = {
                "job_type": "Experiment",
                "job_slug": self.job_slug,
                "in_exp": "true",
                "files": files,
            }
            self.project.sync(**payload)

    def log_artifacts(self, files: list, target=None):
        relative_target = target[1:] if target and target.startswith("/") else target
        try:
            payload = {
                "job_type": "Experiment",
                "workflow_slug": self.job_slug
            }
            self.project.put_files(
                files,
                prefix=relative_target,
                payload=payload,
                message="Log Artifacts Commit",
                tag_images=True
            )
        except APINotFound:  # To support old server version
            payload = {
                "job_type": "Experiment",
                "job": self.job_slug,
                "in_exp": "true",
                "files": files,
            }
            self.project.sync(**payload)

    @suppress_exception
    def restart(self, message=None, sync=True):
        if not in_experiment():
            raise UserError("Cant restart an experiment outside the experiment itself.")
        if sync: self.sync()
        return apis_helper.post(url_join(self._base_url(), 'restart'), data={"message": message})

    @staticmethod
    def get_utilization():
        return spawn_helper.analyze_pid(pid=os.getpid())

    @suppress_exception
    def __latest_artifacts(self, commit=None):
        resp = apis_helper.get(url_join(self._base_url(), 'artifacts'), {"commit": commit})
        if not resp: return None
        return resp.get("artifacts")

    def artifacts(self, commit=None):
        artifacts = self.__latest_artifacts(commit=commit)
        if not artifacts: return []
        return artifacts

    def __wait_until_success(self, fetch_info=60):
        status = None
        is_running = True
        while status != 'Success' and is_running:
            info = self.__get_experiment()
            if info is None:
                return None
            status = info.get("status")
            is_running = info.get("is_running")
            if status == 'Success':
                return {"status": status, "end_commit": info.get("end_commit")}
            if is_running:
                time.sleep(fetch_info)
        return {"status": status}

    @suppress_exception
    def pull_artifacts(self, path=".", commit=None,wait_until_success=False,fetch_info=60):
        if wait_until_success:
            success_resp = self.__wait_until_success(fetch_info=fetch_info)
            if success_resp and success_resp.get("status") == 'Success':
                commit = success_resp.get("end_commit")
            else:
                raise CnvrgError("Experiment has finished with status: {}, can't pull artifacts".format(success_resp.get("status")))

        artifacts = self.artifacts(commit=commit)
        mapped_files = []
        for artifact in artifacts:
            try:
                fpath = os.path.join(path, artifact.get("fullpath"))
                os.makedirs(os.path.dirname(fpath), exist_ok=True)
                artifact_path = apis_helper.download_file(artifact.get("url"), fpath)
                mapped_files.append({**artifact, **{"path": os.path.abspath(artifact_path)}})
            except Exception as e:
                print(e)
                print("Error while downloading {}".format(artifact.get("name")))
        return mapped_files

    @staticmethod
    def run(command=None, arguments: dict=None, grid: dict=None, title: str=None, project: Project=None,
            compute: str=None, output_dir: str=None, computes: List=None,datasets: List=None, dataset: str=None,
            local: bool=False, library: str=None, working_directory: str=None, sync_before: bool=True,
            sync_after: bool=True, prerun: bool=True, requirements: bool=True, image: str=None, schedule: str=None,
            recurring: str=None, git_commit: str=None, git_branch: str=None, local_folders: List=None,
            notify_on_error: bool=False, notify_on_success: bool=False, emails_to_notify: List=None,
            git_diff=False, commit: str=None, **kwargs):
        function_command = None
        project = project or Project()

        if recurring and not croniter.is_valid(recurring):
            raise UserError("Cron is not valid")

        if project.git:
            git_commit = git_commit or project.get_git_commit()
            output_dir = output_dir or "output"
            git_branch = git_branch or project.get_git_branch()
            sync_before = git_diff

        ## sync before
        if sync_before:
            sync_output = project.sync(
                get_output_by_regex='\{.*\"commit_sha1\"\:\".*\"\}',
                **CnvrgJob.current_job_sync_args(output_dir=output_dir, git_diff=git_diff, return_id=True)
            )["output"]
            if len(sync_output) > 0 and project.git:
                sync_output = json.loads(sync_output[0])
                if sync_output["commit_sha1"]:
                    commit = sync_output["commit_sha1"]

        ## support
        computes = computes or wrap_string_to_list(computes) or wrap_string_to_list(compute) or []
        ### by default (if computes empty or no compute) set the local to True
        if local:
            pass
        elif "local" in computes:
            local = True
        elif arguments:
            local = False
        elif grid:
            local = False
        elif len(computes) == 0:
            local = True

        ### if command is function, run it as a function and not as a os command
        if isinstance(command, types.FunctionType): function_command = command

        ## merge datasets
        datasets = datasets or wrap_string_to_list(datasets) or wrap_string_to_list(dataset) or []
        if local:
            Experiment.__verify_local_params(computes, datasets, image, arguments or grid)
            if function_command:
                ## set the title to "func: {name of the function}"
                cmd = "func: {func_name}".format(func_name=function_command.__name__)
                experiment = Experiment.__run_callable(
                    function_command,
                    title=title,
                    project=project,
                    library=library,
                    cmd=cmd,
                    git_commit=git_commit,
                    output_dir=output_dir,
                    git_branch=git_branch,
                    commit=commit
                )
                if not experiment:
                    raise UserError("Cant run experiment, please check your input params")
            else:
                experiment = Experiment.__run_local(
                    cmd=command,
                    title=title,
                    project=project,
                    cwd=working_directory,
                    git_commit=git_commit,
                    git_branch=git_branch,
                    library=library,
                    commit=commit
                )
                if not experiment: raise UserError("Cant run experiment, please check your input params")
        else:
            ## perform an api call to create it.
            Experiment.__verify_remote_params(function_command)
            e = project.run_task(
                command,
                arguments=arguments,
                title=title,
                templates=computes,
                datasets=datasets,
                library=library,
                prerun=prerun,
                requirements=requirements,
                grid=grid,
                git_branch=git_branch,
                git_commit=git_commit,
                output_dir=output_dir,
                image=image,
                schedule=schedule,
                recurring=recurring,
                local_folders=local_folders,
                notify_on_error=notify_on_error,
                notify_on_success=notify_on_success,
                emails_to_notify=emails_to_notify,
                working_directory=working_directory,
                commit=commit,
                **kwargs
            )
            if len(e.get("experiments")) == 1:
                exp = Experiment(e.get("experiments")[0], owner=project.owner)
                exp.notify_running()
                exp.set_as_default()
                return exp
            else:
                hyper_search_url = url_join(project.web_url(),
                                            "experiments?grid={grid}".format(grid=e.get("slug")))
                logger_helper.log_message(
                    "Grid {grid_slug} is running, follow at: {url}".format(grid_slug=e.get("slug"),
                                                                           url=hyper_search_url))
                return [Experiment(exp_id, owner=project.owner) for exp_id in e.get("experiments")]

        if sync_after:
            project.sync(**CnvrgJob.current_job_sync_args(job=experiment['slug'], job_type="Experiment", output_dir=output_dir))
            experiment.__update(end_commit=project.get_current_commit())
        return experiment


    @staticmethod
    def __verify_local_params(computes: List, datasets: List, image: str, arguments: dict):
        if len(computes) > 0 and "local" not in computes:
            logger_helper.log_warn("Computes list: {compute} will be ignored because running on local machine".format(compute=",".join(computes)))
        if len(datasets) > 0:
            logger_helper.log_warn("Dataset list: {datasets} will be ignored because running on local machine".format(datasets=",".join(ds.get("id") for ds in datasets)))
        if image:
            logger_helper.log_warn("Image: {image} will be ignored because running on local".format(image=image))
        if arguments:
            logger_helper.log_warn("Arguments will be ignored because running on local machine")

    @staticmethod
    def __verify_remote_params(function_command):
        if function_command:
            logger_helper.log_warn("Can't run a function command on remote experiment. please use local compute instead.")
            raise UserError("Can't run a function command on remote experiment. please use local compute instead.")

    def log_param(self, key, value=None, run_async=True):
        if not value:
            print("Cant create parameter without value")
            return
        if run_async:
            AsyncRunner(self._log_param).run(key, value)
        else:
            self._log_param(key, value)

    def _log_param(self, key, value=None):
        tag_data = {
            "key": key,
            "value": value,
            "type": TagType.SINGLE_TAG.value
        }
        self.__send_tag(tag_data)

    def __dict__(self):
        return self.__data

    def log_metric(self, key, Ys: List, Xs: List=None, grouping: List=None, x_axis=None, y_axis=None, run_async=True) -> None:
        if run_async:
            AsyncRunner(self._log_metric).run(key, Ys, Xs, grouping, x_axis, y_axis)
        else:
            self._log_metric(key, Ys, Xs, grouping, x_axis, y_axis)

    def _log_metric(self, key, Ys, Xs, grouping: List=None, x_axis=None, y_axis=None) -> None:
        """
        a function which can tag an experiment with a chart
        :param key: the name of the chart
        :param Ys: [y1, y2, y3, y4] (float)
        :param Xs: [x1, x2, x3, x4] (date, integer, null)
        :param grouping: [g1, g2, g3, g4]
        :param x_axis: rename the x_axis of the chart
        :param y_axis:rename the y_axis of the chart
        :return:
        """

        # Support for numpy arrays
        if isinstance(Xs, np.ndarray):
            Xs = Xs.tolist()
        if isinstance(Ys, np.ndarray):
            Ys = Ys.tolist()

        # Support for arrays with numpy values
        if isinstance(Ys, list):
            for index, item in enumerate(Ys):
                if isinstance(item, np.generic):
                    Ys[index] = item.item()
                if isinstance(item, np.ndarray):
                    Ys[index] = item.tolist()

        if isinstance(Xs, list):
            for index, item in enumerate(Xs):
                if isinstance(item, np.generic):
                    Xs[index] = item.item()
                if isinstance(item, np.ndarray):
                    Xs[index] = item.tolist()

        # Support for regular types
        if not isinstance(Ys, list):
            if isinstance(Ys, np.generic):
                Ys = Ys.item()
            Ys = [Ys]

        if Xs is None:
            if key in self.chart_metrics:
                x_stamp = self.chart_metrics[key]
            else:
                self.chart_metrics[key] = 0
                x_stamp = 0
            Xs = [*range(x_stamp, x_stamp + len(Ys))]
            self.chart_metrics[key] += len(Ys)

        if not isinstance(Xs, list):
            if isinstance(Xs, np.generic):
                Xs = Xs.item()
            Xs = [Xs]

        tag_data = {
            "ys": Ys,
            "xs": Xs,
            "key": key,
            "grouping": grouping,
            "x_axis": x_axis,
            "y_axis": y_axis,
            "type": TagType.LINECHART_TAG.value,
        }
        self.__send_tag(tag_data)

    def logs(self, callback=None, poll_every=5):
        job_logs, experiment_is_running = self.__fetch_logs(0)
        offset = len(job_logs)
        callback = callback or logger_helper.log_cnvrg_log
        [callback(l) for l in job_logs]
        while experiment_is_running:
            time.sleep(poll_every)
            job_logs, experiment_is_running = self.__fetch_logs(offset)
            offset += len(job_logs)
            ## filter unknown logs?
            [callback(l) if l else None for l in job_logs]

    @staticmethod
    def list(project, owner=None, flow_slug="", flow_version=""):
        owner_slug = owner or apis_helper.credentials.owner
        if isinstance(project, Project):
            p = project
        else:
            p = Project(project)

        filters = {
            "operator": "AND",
            "conditions": [
                {"key": "job_type", "operator": "is", "value": "Experiment"},
            ]
        }

        flow_query = "{}:{}".format(flow_slug, flow_version)

        if flow_slug:
            condition = {"key": "flow_query", "operator": "is", "value": flow_query}
            filters["conditions"].append(condition)

        resp = apis_helper.get_v2(
            url_join(
                owner_slug,
                "projects",
                p.slug,
                "experiments",
                "?filter={}".format(json.dumps(filters))
            )
        )
        return resp.json()

    def commits(self):
        filter = {
            "operator": "AND",
            "conditions": [{
                "key": "job_id",
                "operator": "is",
                "value": self.job_id()
            }]
        }

        resp = apis_helper.get_v2(url_join(self._base_url(api="v2"), "commits?filter={0}".format(json.dumps(filter))))
        return resp.json()

    def alert(self, message, subject="You have received alert from cnvrg", recipients=None):
        if recipients is None:
            recipients = []
        resp = apis_helper.post_v2(url_join(self._base_url(api="v2"), "alert"), data={
            "message": message,
            "subject": subject,
            "recipients": recipients
        })
        return resp.json()

    @property
    def title(self):
        return self["title"]

    @title.setter
    @suppress_exception
    def title(self, new_title):
        apis_helper.put(url_join(self._base_url(), 'title'), data={"title": new_title})
        self.__data = self.__get_experiment()

    def set_title(self,new_title):
        self.title = new_title

    @suppress_exception
    def show_chart(self, key, **kwargs):
        """

        :param key: chart_key
        :param kwargs: with_legend, legend_loc
        :return:
        """
        chart = apis_helper.get(url_join(self._base_url(), 'charts', key)).get("chart")

        return chart_show_helper.show_chart(chart, **kwargs)

    @suppress_exception
    def __fetch_logs(self, offset, limit=None):
        resp = apis_helper.get(url_join(self._base_url(), 'logs'), data={"offset": offset, "limit": limit})
        return resp.get("logs"), resp.get("experiment").get("is_running")

    def __send_tag(self, tag_data):
        apis_helper.post(url_join(self._base_url(), 'tags'), data={"tag": tag_data})

    @suppress_exception
    def finish(self, exit_status=0):
        apis_helper.post(url_join(self._base_url(), 'finish'), data={"exit_status": exit_status})

    def href(self):
        return self.__data.get("full_href")

    def job_id(self):
        return self.__data.get("job_id")

    def open(self):
        if check_if_lib_exists("webbrowser"):
            import webbrowser
            webbrowser.open(self.href())
        else:
            logger_helper.log_message("You can find the experiment in: {href}".format(href=self.href()))

    def __get_experiment(self):
        resp = apis_helper.get(self._base_url())
        if not resp:
            return None
        if resp.get('status') == 400:
            raise CnvrgError(resp.get('message'))
        return resp.get("experiment")

    def _base_url(self, api="v1"):
        return url_join(
            #### hackish :D
            self.project.get_base_url(api=api),string_helper.to_snake_case(self.job_type) + "s", self.job_slug
        )

    def __getitem__(self, item):
        return self.__data.get(item)

    @suppress_exception
    def __update(self, title: str=None, end_commit: str=None):
        payload = {}
        if title: payload["title"] = title
        if end_commit: payload["end_commit"] = end_commit
        return apis_helper.put(self._base_url(), data={"experiment":payload})



import os

EXPERIMENT = "Experiment"
ENDPOINT = "Endpoint"

ENV_KEYS = {
    "current_job_id": "CNVRG_JOB_ID",
    "current_job_type": "CNVRG_JOB_TYPE",
    "current_project": "CNVRG_PROJECT",
    "current_organization": "CNVRG_OWNER"
}

POOL_SIZE = os.environ.get("CNVRG_THREAD_SIZE") or 20
CURRENT_JOB_ID = os.environ.get("CNVRG_JOB_ID")
CURRENT_JOB_TYPE = os.environ.get("CNVRG_JOB_TYPE")
CNVRG_OUTPUT_DIR = os.environ.get("CNVRG_OUTPUT_DIR")
CURRENT_PROJECT_SLUG = os.environ.get(ENV_KEYS["current_project"])
CURRENT_ORGANIZATION_SLUG = os.environ.get(ENV_KEYS["current_organization"])
MAX_LOGS_PER_SEND = int(os.environ.get("CNVRG_MAX_LOGS_PER_SEND") or 500)

def in_experiment():
    return os.environ.get("CNVRG_JOB_TYPE") == EXPERIMENT

def get_current_job_id():
    return os.environ.get("CNVRG_JOB_ID")

def get_current_job_type():
    return os.environ.get("CNVRG_JOB_TYPE")

def set_current_job_id(job_id):
    os.environ["CNVRG_JOB_ID"] = job_id

def set_current_job_type(job_slug):
    os.environ["CNVRG_JOB_TYPE"] = job_slug

def get_origin_job():
    current_job_id = get_current_job_id()
    current_job_type = get_current_job_type()
    if not current_job_id and not current_job_type: return {}
    return {
        "origin_job_id": current_job_id,
        "origin_job_type": current_job_type
    }


import os

from cnvrg.modules.storage.storage import Storage
from cnvrg.modules.storage.s3.storage import S3Storage
from cnvrg.modules.storage.azure.storage import AzureStorage
from cnvrg.modules.storage.minio.storage import MinioStorage
from cnvrg.modules.storage.gcp.storage import GCPStorage
from cnvrg.modules.cnvrg_files import CnvrgFiles
from cnvrg.helpers.apis_helper import get as apis_get

STORAGE_TYPE_GCP = "gcp"
STORAGE_TYPE_AZURE = "azure"
STORAGE_TYPE_S3 = "s3"
STORAGE_TYPE_MINIO = "minio"


def storage_factory(element: CnvrgFiles, working_dir=None):
    storage_resp = apis_get(element.get_storage_url()).get("client")
    working_dir = working_dir or element.get_working_dir() or os.curdir

    if storage_resp is None:
        return None
    if storage_resp.get("storage") == STORAGE_TYPE_S3:
        return S3Storage(element, working_dir, storage_resp)
    if storage_resp.get("storage") == STORAGE_TYPE_MINIO:
        return MinioStorage(element, working_dir, storage_resp)
    if storage_resp.get("storage") == STORAGE_TYPE_AZURE:
        return AzureStorage(element, working_dir, storage_resp)
    if storage_resp.get("storage") == STORAGE_TYPE_GCP:
        return GCPStorage(element, working_dir, storage_resp)



from functools import wraps
import traceback


def suppress_exception(f):
    @wraps(f)
    def func(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            traceback.print_exc()
    return func



from .base_sql_connector import BaseSqlConnector
is_loaded = False
try:
    import pymysql
    is_loaded = True

except ImportError:
    is_loaded = False



class MysqlConnector(BaseSqlConnector):

    @staticmethod
    def requirements():
        return """pymysql"""

    @staticmethod
    def key_type():
        return "mysql"
    # host – Host where the database server is located
    # user – Username to log in as
    # password – Password to use.
    # database – Database to use, None to not use a particular one.
    # port – MySQL port to use, default is usually OK. (default: 3306)
    # bind_address – When the client has multiple network interfaces, specify the interface from which to connect to the host. Argument can be a hostname or an IP address.
    # unix_socket – Optionally, you can use a unix socket rather than TCP/IP.
    # read_timeout – The timeout for reading from the connection in seconds (default: None - no timeout)
    # write_timeout – The timeout for writing to the connection in seconds (default: None - no timeout)
    # charset – Charset you want to use.
    # sql_mode – Default SQL_MODE to use.
    # read_default_file – Specifies my.cnf file to read these parameters from under the [client] section.
    # conv – Conversion dictionary to use instead of the default one. This is used to provide custom marshalling and unmarshalling of types. See converters.
    # use_unicode – Whether or not to default to unicode strings. This option defaults to true for Py3k.
    # client_flag – Custom flags to send to MySQL. Find potential values in constants.CLIENT.
    # cursorclass – Custom cursor class to use.
    # init_command – Initial SQL statement to run when connection is established.
    # connect_timeout – Timeout before throwing an exception when connecting. (default: 10, min: 1, max: 31536000)
    # ssl – A dict of arguments similar to mysql_ssl_set()’s parameters.
    # read_default_group – Group to read from in the configuration file.
    # compress – Not supported
    # named_pipe – Not supported
    # autocommit – Autocommit mode. None means use server default. (default: False)
    # local_infile – Boolean to enable the use of LOAD DATA LOCAL command. (default: False)
    # max_allowed_packet – Max size of packet sent to server in bytes. (default: 16MB) Only used to limit size of “LOAD LOCAL INFILE” data packet smaller than default (16KB).
    # defer_connect – Don’t explicitly connect on construction - wait for connect call. (default: False)
    # auth_plugin_map – A dict of plugin names to a class that processes that plugin. The class will take the Connection object as the argument to the constructor. The class needs an authenticate method taking an authentication packet as an argument. For the dialog plugin, a prompt(echo, prompt) method can be used (if no authenticate method) for returning a string from the user. (experimental)
    # server_public_key – SHA256 authentication plugin public key value. (default: None)
    # db – Alias for database. (for compatibility to MySQLdb)
    # passwd – Alias for password. (for compatibility to MySQLdb)
    # binary_prefix – Add _binary prefix on bytes and bytearray. (default: False)
    def __init__(self, data_connector):
        super(MysqlConnector, self).__init__(data_connector)

    def _connection(self, username=None, password=None, host=None, **kwargs):
        return pymysql.connect(
            **{"host": host, "password": password, "user": username, **kwargs},cursorclass=pymysql.cursors.DictCursor
        )



import sys


class Progressbar:
    def __init__(self, total, bar_len=100):
        self._total = total
        self._bar_len = bar_len
        self._progress = 0

    @property
    def total(self):
        return self._total

    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, progress):
        if progress > self._total:
            self._progress = self._total
        else:
            self._progress = progress

        self._print_progress()

    def _print_progress(self):
        sys.stdout.write("\r")
        prog_string = ""
        percent = self._progress / self._total
        for i in range(self._bar_len):
            prog_string += "=" if i < int(self._bar_len * percent) else " "
        sys.stdout.write("[ {} ] {}%".format(prog_string, int(percent * 100)))
        sys.stdout.flush()

    def finish(self):
        sys.stdout.write("\n")



import atexit
import threading
import multiprocessing

open_threads = []


class AsyncRunner:
    def __init__(self, func, process_mode=False):
        self.func = func
        self.process_mode = process_mode

        atexit.register(AsyncRunner.wait_fot_threads)

    def run(self, *args, **kwargs):
        thread = threading.Thread(
            target=self.suppressed_func,
            args=args,
            kwargs=kwargs
        )
        thread.start()

        global open_threads
        open_threads.append(thread)

    def suppressed_func(self, *args, **kwargs):
        try:
            self.func(*args, **kwargs)
        except Exception as e:
            print("Caught Cnvrg Exception: {}".format(e))

    @staticmethod
    def wait_fot_threads():
        global open_threads
        for t in open_threads:
            t.join()



from cnvrg.modules.base_module import CnvrgBase
from cnvrg.helpers.apis_helper import update_credentials, credentials as cred


class Cnvrg(CnvrgBase):
    def __init__(self, url="https://app.cnvrg.io", email=None, password=None, owner=None):
        if email and password:
            cred.login(email, password, api_url=url, owner=owner)
            update_credentials(cred)



import os
import boto3
import urllib3

from typing import Dict
from boto3.s3.transfer import TransferConfig
from cnvrg.modules.cnvrg_files import CnvrgFiles
from cnvrg.modules.storage.base_storage import BaseStorage
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

config = TransferConfig(
    multipart_threshold=1024 * 25,
    max_concurrency=10,
    multipart_chunksize=1024 * 25,
    use_threads=True
)


class MinioStorage(BaseStorage):
    def __init__(self, element: CnvrgFiles, working_dir: str, storage_resp: Dict):
        super().__init__(element, working_dir, storage_resp.get("path_sts"))

        try:
            os.remove(self.sts_local_file)
        except Exception:
            pass

        del storage_resp["path_sts"]

        props = self.decrypt_dict(storage_resp, keys=["sts_a", "sts_s", "sts_st", "bucket", "region", "endpoint"])

        self.s3props = {
            "endpoint_url": props.get("endpoint"),
            "aws_access_key_id": props.get("sts_a"),
            "aws_secret_access_key": props.get("sts_s"),
            "region_name": props.get("region")
        }
        self.bucket = props.get("bucket")
        self.region = props.get("region")
        self.client = self._get_client()

    def upload_single_file(self, file, target):
        try:
            self.client.upload_file(file, self.bucket, target, Config=config)
        except Exception as e:
            print(e)

    def download_single_file(self, file, target):
        pass

    def _get_client(self):
        return boto3.client('s3', verify=False, **self.s3props)



import requests
import json
from cnvrg.modules.errors import CnvrgError
import cnvrg.helpers.auth_helper as auth_helper
import cnvrg.helpers.logger_helper as logger_helper
from cnvrg.helpers.url_builder_helper import url_join
import urllib
import ssl
from cnvrg.helpers.error_catcher import suppress_exception
import urllib3

JSON_HEADERS = {
    "Content-Type": "application/json"
}

def verify_logged_in():
    if not credentials.logged_in:
        raise CnvrgError("Not authenticated")


def __get_headers():
    return {"Auth-Token": credentials.token, "Authorization": "CAPI {}".format(credentials.token)}


def __parse_resp(resp, **kwargs):
    try:
        r_j = resp.json()
        if r_j.get("status") and r_j.get("status") != 200 and r_j.get("message"):
            print(r_j.get("message"))
        return r_j
    except Exception as e:
        logger_helper.log_error(resp.text)
        logger_helper.log_bad_response(**kwargs)
        return {"status": 400, "error": str(e)}


@suppress_exception
def request(action, url, **kwargs):
    urllib3.disable_warnings()
    callee = getattr(session, action)
    try:
        resp = callee(url, **kwargs)
        return __parse_resp(resp, url=url, **kwargs)
    except Exception as e:
        return {"status": 400, "error": str(e)}

@suppress_exception
def get_v2(url):
    urllib3.disable_warnings()
    verify_logged_in()
    get_url = url_join(credentials.api_url, "v2", url)
    resp = requests.get(url=get_url, headers=__get_headers(), verify=False)
    if resp.status_code == 404:
        raise CnvrgError('Not Found')
    if resp.status_code != 200:
        messages = []
        for m in resp.json().values():
            messages.append(m) if isinstance(m, str) else messages.append(m[0])
        raise CnvrgError(messages)
    return resp

@suppress_exception
def get(url, data=None):
    verify_logged_in()
    get_url = url_join(base_url, url)
    return request('get', get_url, timeout=420, params=data, verify=False)


@suppress_exception
def post(url, data=None, files=None):
    verify_logged_in()
    get_url = url_join(base_url, url)
    return request('post', get_url, timeout=900, data=json.dumps(data), files=files, verify=False)

@suppress_exception
def post_v2(url, data):
    urllib3.disable_warnings()
    verify_logged_in()
    get_url = url_join(credentials.api_url, "v2", url)
    resp = requests.post(url=get_url, json=data, headers=__get_headers(), verify=False)
    if resp.status_code == 404:
        raise CnvrgError('Not Found')
    if resp.status_code != 200:
        messages = []
        for m in resp.json().values():
            messages.append(m) if isinstance(m, str) else messages.append(m[0])
        raise CnvrgError(messages)
    return resp

@suppress_exception
def put_v2(url, data):
    urllib3.disable_warnings()
    verify_logged_in()
    get_url = url_join(credentials.api_url, "v2", url)
    resp = requests.put(url=get_url, json=data, headers=__get_headers(), verify=False)
    if resp.status_code == 404:
        raise CnvrgError('Not Found')
    if resp.status_code != 200:
        messages = []
        for m in resp.json().values():
            messages.append(m) if isinstance(m, str) else messages.append(m[0])
        raise CnvrgError(messages)
    return resp

@suppress_exception
def send_file(url, data=None, files=None):
    urllib3.disable_warnings()
    verify_logged_in()
    get_url = url_join(base_url, url)
    resp = requests.post(get_url, files=files, data=data, headers={"AUTH-TOKEN": credentials.token, "Authorization": "CAPI {}".format(credentials.token)}, verify=False)
    return __parse_resp(resp, url=url, data=data)


@suppress_exception
def put(url, data=None):
    verify_logged_in()
    get_url = url_join(base_url, url)
    return request('put', get_url, timeout=900, data=json.dumps(data), verify=False)

def download_file(url, fpath):
    try:
        ssl._create_default_https_context = ssl._create_unverified_context
        urllib.request.urlretrieve(url, fpath)
        return fpath
    except Exception as e:
        print(e)

@suppress_exception
def download_raw_file(url):
    resp = requests.get(url, verify=False)
    return resp.content

@suppress_exception
def update_credentials(creds):
    global credentials
    credentials = creds
    global session
    session = requests.session()
    session.headers = {
        "AUTH-TOKEN": credentials.token,
        "Authorization": "CAPI {}".format(credentials.token),
        **JSON_HEADERS
    }
    global base_url
    base_url = url_join(credentials.api_url, "v1")


credentials = auth_helper.CnvrgCredentials()
session = requests.session()
session.headers = {
    "AUTH-TOKEN": credentials.token,
    "Authorization": "CAPI {}".format(credentials.token),
    **JSON_HEADERS
}
base_url = url_join(credentials.api_url, "v1")




from cnvrg.charts import Heatmap, MatrixHeatmap, Scatterplot, Bar, CategoricalScatterplot
try:
    import pandas as pd
    import numpy as np
    pandas_available = True
except ImportError as e:
    pandas_available = False


class PandasAnalyzer(object):
    def __init__(self, data, experiment=None):
        if not pandas_available: raise ImportError("Can't analyze dataframe, please install pandas first.")
        self.__data = data
        self.__categorical_columns = []
        self.__numeric_columns = []
        self.__experiment = experiment
        self.__generate_charts()

    def __generate_charts(self):
        self.__find_correlations()
        self.__categorical_counts()
        self.__null_values()
        self.__correlation_scatter()

    def __categorical_counts(self):
        print("Creating charts of per categorical values.")
        self.__categorical_columns = self.__data.select_dtypes(exclude=["number", "bool_", "object_"]).columns
        for cat_col in self.__categorical_columns:
            counts = self.__data[cat_col].value_counts()
            self.__experiment.log_chart("{}_categorical_counts".format(cat_col),
                                        [Bar(x=counts.index.tolist(), y=counts.values.tolist(), name=cat_col)],
                                        title="Categorical count {col}".format(col=cat_col))

    def __find_correlations(self):
        print("Logging correlation matrix")
        correlation = self.__data.corr()
        self.__experiment.log_chart("correlation", [MatrixHeatmap(np.round(correlation.values, 2))], x_ticks=correlation.index.tolist(), y_ticks=correlation.index.tolist())

    def __null_values(self):
        print("Finding null values")
        null_values = self.__data.isnull().sum()
        self.__experiment.log_chart("null_values", [Bar(x=null_values.index.tolist(), y=null_values.values.tolist())], title="Null values in the dataframe")


    def __correlation_scatter(self):
        indexes = self.__data.select_dtypes(include=["number"]).columns
        corr = self.__data.corr()
        for idx, i in enumerate(indexes):
            for jdx, j in enumerate(indexes):
                if i == j: continue
                if jdx < idx: continue
                corr_val = abs(corr[i][j])
                if 1 ==  corr_val or corr_val < 0.5: continue
                print("create", i, "against", j, "scatter chart")
                droplines = self.__data[[i,j]].notnull().all(1)
                x,y = self.__data[droplines][[i,j]].values.transpose()
                self.__experiment.log_chart("{i}_against_{j}".format(i=i, j=j),
                                            [Scatterplot(x=x.tolist(), y=y.tolist())],
                                            title="{i} against {j}".format(i=i, j=j))








from cnvrg.modules.project import Project
from cnvrg.helpers.url_builder_helper import url_join

import cnvrg.helpers.param_build_helper as param_build_helper
import cnvrg.helpers.apis_helper as apis_helper


class FlowVersion():
    def __init__(self, flow, slug, title, project=None):
        owner, project_slug, flow_slug = param_build_helper.parse_params(flow.slug, param_build_helper.FLOW)

        p = Project.factory(owner, project)

        if p is None:
            p = Project(url_join(owner, project_slug))

        self.slug = slug
        self.title = title
        self.owner = owner
        self.flow = flow
        self.project = p

    def info(self):
        resp = apis_helper.get_v2(
            url_join(
                self.project.get_base_url(api="v2"),
                'flows',
                self.flow.slug,
                "flow-versions",
                self.slug,
                'info'
            )
        )

        return resp.json()["data"]["attributes"].get("fv_status")



# updated_on_server, updated_on_local, conflicts, deleted_conflicts, deleted_on_local, deleted_on_remote
class CommitStatus:
    UPDATED_ON_SERVER = "updated_on_server"
    UPDATED_ON_LOCAL = "updated_on_local"
    CONFLICTS = "conflicts"
    DELETED_CONFLICTS = "deleted_conflicts"
    DELETED_ON_LOCAL = "deleted_on_local"
    DELETED_ON_REMOTE = "deleted_on_remote"
    COMMIT_SHA1 = "commit_sha1"



from setuptools import setup, find_packages
from cnvrg.version import VERSION
setup(
    name='cnvrg',
    version=VERSION,
    python_requires='>3.5',
    packages=find_packages(exclude=("tests",)),
    install_requires=[
        'click',
        'requests==2.23.0',
        "boto3",
        'colorama',
        "tqdm",
        "pyaml",
        'tinynetrc',
        'pycryptodome',
        'psutil',
        'dill',
        'pytz',
        'numpy',
        'pandas',
        'azure-core==1.10.0',
        'azure-storage-blob==12.6.0',
        'google-cloud-storage',
        'six>=1.13',
        'progress',
        'croniter'
    ],
    author="cnvrg",
    author_email="support@cnvrg.io",
    entry_points='''
        [console_scripts]
        cnvrgp=cnvrg.main:cli
    ''',
)



from cnvrg.modules.errors import CnvrgError

def args_to_string(args):
    if args == None: return ''
    if isinstance(args, list): args = {k["key"]: k.get("value") for k in args}
    ### expect dict of key=value
    return " ".join(map(lambda x: k_v_arg_to_string(x[0], x[1]), args.items()))

def k_v_arg_to_string(k,v):
    if v == None:
        return "--{key}".format(key=k)
    elif isinstance(v, bool) and v:
        return "--{key}".format(key=k)
    elif isinstance(v, bool) and not v:
        return "--{key}=false".format(key=k)
    elif isinstance(v, float) or isinstance(v, int):
        return "--{key}={value}".format(key=k, value=v)
    else:
        return "--{key}='{value}'".format(key=k, value=v)


def validate_args(args):
    for key,value in args.items():
        if isinstance(value, list):
            raise CnvrgError("Library local run with many arguments are not supported yet.")


from cnvrg.subcommands.library import library as lib_cli
from cnvrg.subcommands.project import project as project_cli
from cnvrg.subcommands.cnvrg import cnvrg as cnvrg_cli
import click




cli = click.CommandCollection(sources=[lib_cli, project_cli, cnvrg_cli])



if __name__ == '__main__':
    cli()
















